@@ -1,7 +1,7 @@
 package App::lcpan;
 
-our $DATE = '2015-04-28'; # DATE
-our $VERSION = '0.31'; # VERSION
+our $DATE = '2015-04-29'; # DATE
+our $VERSION = '0.32'; # VERSION
 
 use 5.010001;
 use strict;
@@ -186,7 +186,7 @@ sub _relpath {
 }
 
 our $db_schema_spec = {
-    latest_v => 5,
+    latest_v => 6,
 
     install => [
         'CREATE TABLE author (
@@ -226,6 +226,14 @@ our $db_schema_spec = {
         'CREATE INDEX ix_module__file_id ON module(file_id)',
         'CREATE INDEX ix_module__cpanid ON module(cpanid)',
 
+        'CREATE TABLE namespace (
+            name VARCHAR(255) NOT NULL,
+            num_sep INT NOT NULL,
+            has_child BOOL NOT NULL,
+            num_modules INT NOT NULL
+        )',
+        'CREATE UNIQUE INDEX ix_namespace__name ON namespace(name)',
+
         'CREATE TABLE dist (
              id INTEGER NOT NULL PRIMARY KEY,
              name VARCHAR(90) NOT NULL,
@@ -318,6 +326,38 @@ our $db_schema_spec = {
         'ALTER TABLE dist ADD COLUMN is_latest BOOLEAN',
     ],
 
+    upgrade_to_v6 => [
+        'CREATE TABLE namespace (
+            name VARCHAR(255) NOT NULL,
+            num_sep INT NOT NULL,
+            has_child BOOL NOT NULL,
+            num_modules INT NOT NULL
+        )',
+        'CREATE UNIQUE INDEX ix_namespace__name ON namespace(name)',
+        sub {
+            my $dbh = shift;
+            my $sth_sel_mod = $dbh->prepare("SELECT name FROM module");
+            my $sth_ins_ns  = $dbh->prepare("INSERT INTO namespace (name, num_sep, has_child, num_modules) VALUES (?,?,?,1)");
+            my $sth_upd_ns_inc_num_mod = $dbh->prepare("UPDATE namespace SET num_modules=num_modules+1, has_child=1 WHERE name=?");
+            $sth_sel_mod->execute;
+            my %cache;
+            while (my ($mod) = $sth_sel_mod->fetchrow_array) {
+                my $has_child = 0;
+                while (1) {
+                    if ($cache{$mod}++) {
+                        $sth_upd_ns_inc_num_mod->execute($mod);
+                    } else {
+                        my $num_sep = 0;
+                        while ($mod =~ /::/g) { $num_sep++ }
+                        $sth_ins_ns->execute($mod, $has_child, $num_sep);
+                    }
+                    $mod =~ s/::\w+\z// or last;
+                    $has_child = 1;
+                }
+            }
+        },
+    ],
+
     # for testing
     install_v1 => [
         'CREATE TABLE author (
@@ -662,9 +702,27 @@ sub _update_index {
                 push @old_filenames, $fname;
             }
             last CLEANUP unless @old_file_ids;
-            $log->tracef("  Deleting old files: %s", \@old_filenames);
+            $log->tracef("  Deleting old files records: %s", \@old_filenames);
             $dbh->do("DELETE FROM dep WHERE file_id IN (".join(",",@old_file_ids)."))");
-            $dbh->do("DELETE FROM module WHERE file_id IN (".join(",",@old_file_ids).")");
+            {
+                my $sth = $dbh->prepare("SELECT name FROM module WHERE file_id IN (".join(",",@old_file_ids).")");
+                $sth->execute;
+                my @mods;
+                while (my ($mod) = $sth->fetchrow_array) {
+                    push @mods, $mod;
+                }
+
+                my $sth_upd_ns_dec_num_mod = $dbh->prepare("UPDATE namespace SET num_modules=num_modules-1 WHERE name=?");
+                for my $mod (@mods) {
+                    while (1) {
+                        $sth_upd_ns_dec_num_mod->execute($mod);
+                        $mod =~ s/::\w+\z// or last;
+                    }
+                }
+                $dbh->do("DELETE FROM namespace WHERE num_modules <= 0");
+
+                $dbh->do("DELETE FROM module WHERE file_id IN (".join(",",@old_file_ids).")");
+            }
             {
                 my $sth = $dbh->prepare("SELECT name FROM dist WHERE file_id IN (".join(",",@old_file_ids).")");
                 $sth->execute;
@@ -1006,6 +1064,7 @@ sub stats {
 
     ($stat->{num_authors}) = $dbh->selectrow_array("SELECT COUNT(*) FROM author");
     ($stat->{num_modules}) = $dbh->selectrow_array("SELECT COUNT(*) FROM module");
+    ($stat->{num_namespaces}) = $dbh->selectrow_array("SELECT COUNT(*) FROM namespace");
     ($stat->{num_dists}) = $dbh->selectrow_array("SELECT COUNT(DISTINCT name) FROM dist");
     (
         $stat->{num_releases},
@@ -1242,6 +1301,11 @@ $SPEC{modules} = {
         %fauthor_args,
         %fdist_args,
         %flatest_args,
+        namespace => {
+            summary => 'Select modules belonging to certain namespace',
+            schema => 'str*',
+            tags => ['category:filtering'],
+        },
     },
     result => {
         description => <<'_',
@@ -1281,6 +1345,11 @@ sub modules {
         push @where, "(dist=?)";
         push @bind, $args{dist};
     }
+    if ($args{namespace}) {
+        return [400, "Invalid namespace, please use Word or Word(::Sub)+"]
+            unless $args{namespace} =~ /\A\w+(::\w+)*\z/;
+        push @where, "(name='$args{namespace}' OR name LIKE '$args{namespace}%')";
+    }
     if ($args{latest}) {
         push @where, "(SELECT is_latest FROM dist d WHERE d.file_id=module.file_id)";
     } elsif (defined $args{latest}) {
@@ -1896,6 +1965,86 @@ sub rdeps {
     $res;
 }
 
+$SPEC{namespaces} = {
+    v => 1.1,
+    summary => 'List namespaces',
+    args => {
+        %common_args,
+        %query_args,
+        from_level => {
+            schema => ['int*', min=>0],
+            tags => ['category:filtering'],
+        },
+        to_level => {
+            schema => ['int*', min=>0],
+            tags => ['category:filtering'],
+        },
+        level => {
+            schema => ['int*', min=>0],
+            cmdline_aliases => {l=>{}},
+            tags => ['category:filtering'],
+        },
+        sort => {
+            schema => ['str*', in=>[qw/name -name num_modules -num_modules/]],
+            tags => ['category:sorting'],
+        },
+    },
+};
+sub namespaces {
+    my %args = @_;
+
+    _set_args_default(\%args);
+    my $cpan = $args{cpan};
+    my $index_name = $args{index_name};
+    my $detail = $args{detail};
+    my $q = $args{query} // ''; # sqlite is case-insensitive by default, yay
+    $q = '%'.$q.'%' unless $q =~ /%/;
+
+    my $dbh = _connect_db('ro', $cpan, $index_name);
+
+    my @bind;
+    my @where;
+    if (length($q)) {
+        push @where, "(name LIKE ?)";
+        push @bind, $q;
+    }
+    if (defined $args{from_level}) {
+        push @where, "(num_sep >= ?)";
+        push @bind, $args{from_level}-1;
+    }
+    if (defined $args{to_level}) {
+        push @where, "(num_sep <= ?)";
+        push @bind, $args{to_level}-1;
+    }
+    if (defined $args{level}) {
+        push @where, "(num_sep = ?)";
+        push @bind, $args{level}-1;
+    }
+    my $order = 'name';
+    if ($args{sort} eq 'num_modules') {
+        $order = "num_modules";
+    } elsif ($args{sort} eq '-num_modules') {
+        $order = "num_modules DESC";
+    }
+    my $sql = "SELECT
+  name,
+  num_modules
+FROM namespace".
+    (@where ? " WHERE ".join(" AND ", @where) : "")."
+ORDER BY $order";
+
+    my @res;
+    my $sth = $dbh->prepare($sql);
+    $sth->execute(@bind);
+    while (my $row = $sth->fetchrow_hashref) {
+        push @res, $detail ? $row : $row->{name};
+    }
+    my $resmeta = {};
+    $resmeta->{format_options} = {any=>{table_column_orders=>[[qw/name num_modules/]]}}
+        if $detail;
+    [200, "OK", \@res, $resmeta];
+}
+
 1;
 # ABSTRACT: Manage your local CPAN mirror
 
@@ -1911,7 +2060,7 @@ App::lcpan - Manage your local CPAN mirror
 
 =head1 VERSION
 
-This document describes version 0.31 of App::lcpan (from Perl distribution App-lcpan), released on 2015-04-28.
+This document describes version 0.32 of App::lcpan (from Perl distribution App-lcpan), released on 2015-04-29.
 
 =head1 SYNOPSIS
 
@@ -2157,6 +2306,10 @@ Filename of index.
 
 =item * B<latest> => I<bool>
 
+=item * B<namespace> => I<str>
+
+Select modules belonging to certain namespace.
+
 =item * B<query> => I<str>
 
 Search query.
@@ -2179,6 +2332,52 @@ By default will return an array of package names. If you set C<detail> to true,
 will return array of records.
 
 
+=head2 namespaces(%args) -> [status, msg, result, meta]
+
+List namespaces.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<cpan> => I<str>
+
+Location of your local CPAN mirror, e.g. /path/to/cpan.
+
+Defaults to C<~/cpan>.
+
+=item * B<detail> => I<bool>
+
+=item * B<from_level> => I<int>
+
+=item * B<index_name> => I<str> (default: "index.db")
+
+Filename of index.
+
+=item * B<level> => I<int>
+
+=item * B<query> => I<str>
+
+Search query.
+
+=item * B<sort> => I<str>
+
+=item * B<to_level> => I<int>
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
 =head2 packages(%args) -> [status, msg, result, meta]
 
 List modules/packages.
@@ -2209,6 +2408,10 @@ Filename of index.
 
 =item * B<latest> => I<bool>
 
+=item * B<namespace> => I<str>
+
+Select modules belonging to certain namespace.
+
 =item * B<query> => I<str>
 
 Search query.
